==> kom%2B%2B-0.1.tar.gz: kom++-0.1/tcl-clients/nilkom_cmd.tcl <==
# Command interpreter for nilkom

# global variables (local to this file)
#
#  nilkom_prompt - the current prompt. Determines the next command to use.
#                  Values beginnig with NK_ belong entirely to the TCL layer,
#                  other values originate from [kom_unread prompt].

# global variables (shared)
#  review_list   - list of textnos to review

# The command table. Each entry consists of command template and a
# command template.

set cmdtab {
    {{se tiden} 
	{se tiden}}
    {{markera}
	{mark $senaste 127}}
    {{markera %INT}
	{mark %0 127}}
    {{markera %INT %INT} 
	{mark %1 %0}}
    {{terse senaste} 		
	{print_text $senaste}}
    {{terse %INT} 		
	{print_text %0}}
    {{terse text %INT} 
	{print_text %0}}
    {{nsta mte} 	
	{next_conference}}
    {{kommentera} 
	{k}}
    {{inlgg}
	{start_write_text}
	{(skriva) inlgg}}
    {{fortstt}
	{create_text_resume_edit}
	{fortstt (editera inlgget)}}
    {{hela} 
	{create_text_redisplay; print {\n}}}
    {{terse alla kommentarer}		
	{review_all_comments $senaste}}
    {{terse alla kommentarer till %INT} 
	{review_all_comments %0}}
    {{terse kommenterade} 		
	{review_commented $senaste}
	{terse (det) kommenterade}}
    {{terse det som kommenteras av %INT}
	{review_commented %0}}
    {{skicka} 
	{enter_text}}
    {{bort} 
	{throw_text}}
    {{lsmarkera inte}
	{reject} {lsmarkera inte (senast lsta text)}}
    {{lista kommandon}
	{list_commands}}
    {{lista nyheter}
	{unread}
	{lista (alla) nyheter}}
    {{hjlp}
	{help}}
    {{?}
	{help}}
    {{help}
	{help}}
}

proc list_commands {} {
    global cmdtab

    foreach cmd $cmdtab {
	if {[llength $cmd] > 2} {
	    print "[lindex $cmd 2]\n"
	} else {
	    print "[lindex $cmd 0]\n"
	}
    }
}

proc print_prompt {} {
    global nilkom_prompt
    global review_list

    trim_review_list
    if {[llength $review_list] > 0} {
	set nilkom_prompt NK_REVIEW
    } else {
	set nilkom_prompt [kom_unread prompt]
    }

    case $nilkom_prompt in {
	NEXT_CONF     {print "G till nsta mte - "}
	NEXT_TEXT     {print "Lsa nsta text - "}
	NEXT_COMMENT  {print "Lsa nsta kommentar - "}
	NEXT_FOOTNOTE {print "Lsa nsta fotnot - "}
	NO_MORE_TEXT  {print "Se tiden - "}
	NK_REVIEW     {print "terse nsta - "}
    }
    flush stdout
}

proc default_action {} {
    global nilkom_prompt
    global to_mark_as_read

    foreach tno $to_mark_as_read {
	kom_text mark_as_read $tno
    }
    set to_mark_as_read {}

    case $nilkom_prompt in {
      NEXT_CONF     {print [format "Du r nu nrvarande i %s\n" \
			     [kom_conference name \
					 [kom_unread goto_next_conf]]]}
      NEXT_TEXT     {read_text [kom_unread next_text remove]}
      NEXT_COMMENT  {read_text [kom_unread next_text remove]}
      NEXT_FOOTNOTE {read_text [kom_unread next_text remove]}
      NO_MORE_TEXT  {se tiden}
      NK_REVIEW     {review_next}
    }
}

proc isoify {str} {
    regsub -all \{ $str  str
    regsub -all \} $str  str
    regsub -all \\\| $str  str
    regsub -all \\\[ $str  str
    regsub -all \\\] $str  str
    regsub -all \\\\ $str  str
    return $str
}

# Entry:
# list of {0: cmd-name 1: cmd-tmpl 2: list of {0: mcp 1: data}}
# Exit:
# list of {0: cmd-name 1: exec-cmd}

proc expand_matches {input} {
    set res {}
    set bestmcp 1000
    set bestlwp 1000
    foreach cmd $input {
	set cmdname [lindex $cmd 0]
	set cmdtmpl [lindex $cmd 1]
	foreach matchalt [lindex $cmd 2] {
	    set execcmd {}
	    set mcp [lindex $matchalt 0]
	    set lwp 0
	    set data [lindex $matchalt 1]
	    foreach word $cmdtmpl {
		if {[string range $word 0 0] == "%"} {
		    # Hr behver vi sl upp saker i servern, ibland.
		    # Nr vi gr det hr kan lwp och/eller mcp ka 
		    # (inte om vi slr upp en %CONFLIST, men om vi
		    # slr upp en %CONFUNIQ).
		    append execcmd " [lindex $data [string range $word 1 end]]"
		} else {
		    append execcmd " $word"
		}
	    }
	    if {$lwp < $bestlwp} {
		set res {}
		set bestlwp $lwp
		set bestmcp 1000
	    }
	    if {$lwp <= $bestlwp} {
		if {$mcp < $bestmcp} {
		    set res {}
		    set bestmcp {}
		}
		if {$mcp <= $bestmcp} {
		    lappend res [list $cmdname $execcmd]
		}
	    }
	}
    }
    return $res
}


# Penalties:
#  * For each word with at least one missing character, add 1.
#  * For each conference name with a missing word, add 10.
#  * For each command word missing, add 100.
# If noise is given, fail and return -1.

# lpf - Lowest cwp Penalty Found
# trs - Try ReSult 0: cwp. 1: list of matchdata: {0: mcp: 1 data}
#  Note: the mcp must be returned, since the lwp dominates it.
# cwp - command word penalty
# lwp - lookup word penalty
# mcp - missing character penalty

proc anchor_left_try {try pattern cwp mcp lpf} {
    # puts "-> $try <-> $pattern <- CWP $cwp MCP $mcp LPF $lpf"
    if {[llength $try] == 0} {
	foreach p $pattern {
	    if {[string range $p 0 0] == "%"} {
		# %-sequences must much!
		return [list -1 {}]
	    }
	    incr cwp
	}
	set foo [list $cwp [list $mcp {}]]
	# puts "--->| $foo"
	return $foo
    }
    
    set next_try [lrange $try 1 end]
    set t [lindex $try 0]
    set good_matches {}
    
    # Loop through all patterns, but prune the search tree if
    # the penalty so far exceeds the lowest penalty found.
    for {set anchor 0} {$anchor < [llength $pattern] && $cwp <= $lpf} {incr anchor} {
	set p [lindex $pattern $anchor]
	if {[string range $p 0 0] == "%"} {
	    set trs [try_special $try [lrange $pattern $anchor end] \
			 $cwp $mcp $lpf]
	} else {
	    # Normal word matching.
	    if {[string first $t $p] == 0} {
		# They matched!
		set trs \
		    [anchor_left_try \
		     $next_try \
			 [lrange $pattern [expr {$anchor + 1}] end] \
			 $cwp \
			 [expr {$mcp + ([string length $t] \
					    < [string length $p])}] \
			 $lpf]
		# puts "Got $trs"
	    } else {
		set trs [list -1 {}]
	    }
	}
	
	if {[lindex $trs 0] >= 0 && [lindex $trs 0] <= $lpf} {
	    # A good match found!
	    # puts "MATCH: $try <-> $pattern (CWP [lindex $trs 0])"
	    if {[lindex $trs 0] < $lpf} {
		# This is the best match so far!
		set lpf [lindex $trs 0]
		set good_matches {}
	    }
	    set good_matches [concat $good_matches [lindex $trs 1]]
	}
	
	incr cwp
    }

    # puts "---<--- [list $lpf $good_matches]"
    return [list $lpf $good_matches]
}

proc try_special {try pattern cwp mcp lpf} {
    set data {}
    switch [string range [lindex $pattern 0] 1 end] {
	INT
	{
	    if {[regexp {^[0-9]*$} [lindex $try 0]] == 0} {
		return -1
	    }
	    set trs [anchor_left_try \
			 [lrange $try 1 end] \
			 [lrange $pattern 1 end] \
			 $cwp $mcp $lpf]
	    set ncwp [lindex $trs 0]
	    set nmd {}
	    foreach d [lindex $trs 1] {
		lappend nmd [list [lindex $d 0] \
				 [concat [lindex $d 1] \
				      [list [lindex $try 0]]]]
	    }
	    return [list $ncwp $nmd]
	}
	default
	{
	    error "Bad operand type: $p"
	}
    }
    return {-1 {}}
}

# Return value:
#  0: cwp 1: cmd-name 2: cmd-tmpl 3: list of matchdata: {0: mcp 1: data}
# Note: the mcp field is likely to increase if a name lookup is done.
proc try_match {try entry} {
    set args {}
    set pattern [lindex $entry 0]
    set try [split $try " \t"]
    set trs [anchor_left_try $try $pattern 0 0 1000]
    if {[lindex $trs 0] == 1000} {
	set trs {-1 {}}
    } else {
	set trs [list [lindex $trs 0] \
		     [lindex $entry [expr {[llength $entry] > 2 ? 2 : 0}]] \
		     [lindex $entry 1] \
		     [lindex $trs 1]]
    }
    # puts "try_match returning $trs"
    return $trs
}


proc do_command {cmd} {
    global cmdtab
    global errorCode

    if {[string length $cmd] == 0} {
	perform_command {Foo {default_action}}
	return
    }

    if {[string range $cmd 0 0] == "="} {
	if {[catch {uplevel \#0 [string range $cmd 1 end]} res] == 0} {
	    if {[string length $res] > 0} {
	        print "$res\n"
	    }
	} else {
	    puts "Error: $res"
	    puts "errorCode: $errorCode"
	}
	return
    }
	    
    set cmd [isoify $cmd]

    set best 1000
    set matches {}

    foreach entry $cmdtab {
	set score [try_match $cmd $entry]
	set s [lindex $score 0]
	if {$s != -1} {
	    if {$s < $best} {
		set best [lindex $score 0]
		set matches [list [lrange $score 1 end]]
	    } else {
		if {$s == $best} {
		    lappend matches [lrange $score 1 end]
		}
	    }
	}
    }

    # Here, matches looks like this:
    # list of {0: cmd-name 1: cmd-tmpl 2: list of {0: mcp 1: data}}

    # puts "Ok, matches so far: $matches (End matches)"

    set matches [expand_matches $matches]

    # puts "Ok, matches now: $matches (End matches)"

    switch [llength $matches] {
	0
	    {print "Jag frstr inte vad du menar\n"}
	1
	    {perform_command [lindex $matches 0]}
	default
	    {report_ambiguous $matches}
    }
}

proc perform_command {cmd} {
    global senaste
    global errorCode

    # puts "DOING-->[lindex $cmd 1]<--"
    if {[catch {eval [lindex $cmd 1]} res] != 0} {
	puts "Error: $res"
	puts "errorCode: $errorCode"
    }
}

proc report_ambiguous {alts} {
    print "Du kan mena ngot av fljande:\n"
    foreach m $alts {
        print "[lindex $m 0]\n"
    }
}

proc commandline {n} {
    global input_line
    global inhibit_prompt

    set inhibit_prompt 0

    switch $n {
	3
	    {return 1}
	10
	    -
	13  
	    {
		print "\n"
		do_command $input_line
	     	set input_line {}
		if {! $inhibit_prompt} {
	     	    print_prompt
		}
	    }
	8
	    -
	127
	    {
		set len [string length $input_line]
		if {$len == 0} {
		    print "\a"
		} else {
		    set input_line \
			[string range $input_line 0 [expr $len - 2]]
		    print "\008 \008"
		}
	    }
	18
	    {print "\n$input_line"}
	21
	    {
		print "\n"
		set input_line {}
	    }
	default
	    {
		print [format "%c" $n]
		append input_line [format "%c" $n]
	    }
    }
    flush stdout
    return 0
}

==> kom%2B%2B-0.2.tar.gz: kom++-0.2/tcl-clients/nilkom_cmd.tcl <==
# The command interpreter for the nilkom client.
#
# $Id: nilkom_cmd.tcl,v 1.20 1994/07/16 19:30:35 ceder Exp $
# Copyright (C) 1994  Per Cederqvist and Inge Wallin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


# ----------------------------------------------------------------
# global variables (local to this file)
#
#  nilkom_prompt - the current prompt. Determines the next command to use.
#                  Values beginnig with NK_ belong entirely to the TCL layer,
#                  other values originate from [kom_unread prompt].

# global variables (shared)
#  review_list   - list of textnos to review

# The command table. Each entry consists of 
#    - a user command template
#    - a TCL command template
#    - a pretty-printing user command name (optional)
#
# The user command template is a list of words.  The input the user
# enters must match the words in the user command template.  The
# template can contain argument placeholders.  The take one of the
# following forms:
#    %INT   - An integer.
# In the future, more forms will be supported.  Example: %CONFUNIQ
# will mean a unique conference name.
#
# If the match routine can find a best-match from the table, the TCL
# command template will be used to form a TCL command to execute.  In
# its simplest form, a TCL command template is a TCL command to
# execute.  If the User command template contained a placeholder, the
# corresponding user-supplied argument is inserted when a word
# consisting of %<digit> is encountered.  %0 is the leftmost argument,
# %1 the next, and so on.
#
# The pretty-printing user command name is used by the help commands.

set cmdtab {
    {{vilka}
	{who_is_on_small}
	{vilka (r inloggade)}}
    {{lng vilkalista}
	{who_is_on_verbose}
	{(visa) lng vilkalista}}
    {{se tiden} 
	{se tiden}}
    {{sluta}
	{quit_command}}
    {{markera}
	{mark $senaste 127}
	{markera (senast lsta med markering 127)}}
    {{markera %INT}
	{mark %0 127}
	{markera <inlggsnummer> (med markering 127)}}
    {{markera %INT %INT} 
	{mark %1 %0}
	{markera <inlggsnummer> (med markering) <markeringsnummer>}}
    {{terse senaste} 		
	{print_text $senaste}}
    {{terse %INT} 		
	{print_text %0}
	{terse <inlggsnummer>}}
    {{terse text %INT} 
	{print_text %0}
	{terse text <inlggsnummer>}}
    {{nsta mte} 	
	{next_conference}}
    {{kommentera} 
	{k}}
    {{kommentera %INT} 
	{k %0}
	{kommentera <inlggsnummer>}}
    {{personligt}
	{start_personal_answer $senaste}
	{personligt (svar p senaste inlgget)}}
    {{personligt %INT}
	{start_personal_answer %0}
	{personligt (svar p inlgg) <inlggsnummer>}}
    {{inlgg}
	{start_write_text}
	{(skriva) inlgg}}
    {{fortstt}
	{create_text_resume_edit}
	{fortstt (editera inlgget)}}
    {{hela} 
	{create_text_redisplay; print {\n}}
	{(visa) hela (inlgget som editeras)}}
    {{terse alla kommentarer}		
	{review_all_comments $senaste}}
    {{terse alla kommentarer %INT}
	{review_all_comments %0}
	{terse alla kommentarer (till) <inlggsnumemr>}}
    {{terse kommenterade} 		
	{review_commented $senaste}
	{terse (det) kommenterade}}
    {{terse kommenteras av %INT}
	{review_commented %0}
	{terse (det som) kommenteras av <inlggsnummer>}}
    {{skicka} 
	{enter_text}
	{skicka (in texten)}}
    {{bort} 
	{throw_text}
	{(kasta) bort (inlgget som editeras)}}
    {{lsmarkera inte}
	{reject}
	{lsmarkera inte (senast lsta text)}}
    {{lista kommandon}
	{list_commands}
	{lista (alla) kommandon}}
    {{lista nyheter}
	{unread}
	{lista (alla) nyheter}}
    {{hjlp}
	{help}}
    {{?}
	{help}
	{(Vad kan jag gra)?}}
    {{help}
	{help}}
}

proc list_commands {} {
    global cmdtab

    foreach cmd $cmdtab {
	if {[llength $cmd] > 2} {
	    print "[lindex $cmd 2]\n"
	} else {
	    print "[lindex $cmd 0]\n"
	}
    }
}

proc print_prompt {} {
    global nilkom_prompt
    global review_list

    trim_review_list
    if {[llength $review_list] > 0} {
	set nilkom_prompt NK_REVIEW
    } else {
	set nilkom_prompt [kom_unread prompt]
    }

    case $nilkom_prompt in {
	NEXT_CONF     {print "G till nsta mte - "}
	NEXT_TEXT     {print "Lsa nsta text - "}
	NEXT_COMMENT  {print "Lsa nsta kommentar - "}
	NEXT_FOOTNOTE {print "Lsa nsta fotnot - "}
	NO_MORE_TEXT  {print "Se tiden - "}
	NK_REVIEW     {print "terse nsta - "}
    }
    flush stdout
    nilkom_page sync
}

proc default_action {} {
    global nilkom_prompt
    global to_mark_as_read

    foreach tno $to_mark_as_read {
	kom_text mark_as_read $tno
    }
    set to_mark_as_read {}

    case $nilkom_prompt in {
      NEXT_CONF     {set new_conf [kom_unread goto_next_conf]; \
		     print [format "Du hamnade i %s (%d olsta)\n" \
				[kom_conference name $new_conf] \
				[lindex [kom_unread unread_texts_in_current] \
				        1]]}
      NEXT_TEXT     {read_text [kom_unread next_text remove]}
      NEXT_COMMENT  {read_text [kom_unread next_text remove]}
      NEXT_FOOTNOTE {read_text [kom_unread next_text remove]}
      NO_MORE_TEXT  {se tiden}
      NK_REVIEW     {review_next}
    }
}

proc isoify {str} {
    regsub -all \{ $str  str
    regsub -all \} $str  str
    regsub -all \\\| $str  str
    regsub -all \\\[ $str  str
    regsub -all \\\] $str  str
    regsub -all \\\\ $str  str
    return $str
}

# Entry:
# list of {0: cmd-name 1: cmd-tmpl 2: list of {0: mcp 1: data}}
# Exit:
# list of {0: cmd-name 1: exec-cmd}

proc expand_matches {input} {
    set res {}
    set bestmcp 1000
    set bestlwp 1000
    foreach cmd $input {
	set cmdname [lindex $cmd 0]
	set cmdtmpl [lindex $cmd 1]
	foreach matchalt [lindex $cmd 2] {
	    set execcmd {}
	    set mcp [lindex $matchalt 0]
	    set lwp 0
	    set data [lindex $matchalt 1]
	    foreach word $cmdtmpl {
		if {[string range $word 0 0] == "%"} {
		    # Hr behver vi sl upp saker i servern, ibland.
		    # Nr vi gr det hr kan lwp och/eller mcp ka 
		    # (inte om vi slr upp en %CONFLIST, men om vi
		    # slr upp en %CONFUNIQ).
		    append execcmd " [lindex $data [string range $word 1 end]]"
		} else {
		    append execcmd " $word"
		}
	    }
	    if {$lwp < $bestlwp} {
		set res {}
		set bestlwp $lwp
		set bestmcp 1000
	    }
	    if {$lwp <= $bestlwp} {
		if {$mcp < $bestmcp} {
		    set res {}
		    set bestmcp {}
		}
		if {$mcp <= $bestmcp} {
		    lappend res [list $cmdname $execcmd]
		}
	    }
	}
    }
    return $res
}


# Penalties:
#  * For each word with at least one missing character, add 1.
#  * For each conference name with a missing word, add 10.
#  * For each command word missing, add 100.
# If noise is given, fail and return -1.

# lpf - Lowest cwp Penalty Found
# trs - Try ReSult 0: cwp. 1: list of matchdata: {0: mcp: 1 data}
#  Note: the mcp must be returned, since the lwp dominates it.
# cwp - command word penalty
# lwp - lookup word penalty
# mcp - missing character penalty

proc anchor_left_try {try pattern cwp mcp lpf} {
    # puts "-> $try <-> $pattern <- CWP $cwp MCP $mcp LPF $lpf"
    if {[llength $try] == 0} {
	foreach p $pattern {
	    if {[string range $p 0 0] == "%"} {
		# %-sequences must much!
		return [list -1 {}]
	    }
	    incr cwp
	}
	set foo [list $cwp [list $mcp {}]]
	# puts "--->| $foo"
	return $foo
    }
    
    set next_try [lrange $try 1 end]
    set t [lindex $try 0]
    set good_matches {}
    
    # Loop through all patterns, but prune the search tree if
    # the penalty so far exceeds the lowest penalty found.
    for {set anchor 0} {$anchor < [llength $pattern] && $cwp <= $lpf} {incr anchor} {
	set p [lindex $pattern $anchor]
	if {[string range $p 0 0] == "%"} {
	    set trs [try_special $try [lrange $pattern $anchor end] \
			 $cwp $mcp $lpf]
	} else {
	    # Normal word matching.
	    if {[string first $t $p] == 0} {
		# They matched!
		set trs \
		    [anchor_left_try \
		     $next_try \
			 [lrange $pattern [expr {$anchor + 1}] end] \
			 $cwp \
			 [expr {$mcp + ([string length $t] \
					    < [string length $p])}] \
			 $lpf]
		# puts "Got $trs"
	    } else {
		set trs [list -1 {}]
	    }
	}
	
	if {[lindex $trs 0] >= 0 && [lindex $trs 0] <= $lpf} {
	    # A good match found!
	    # puts "MATCH: $try <-> $pattern (CWP [lindex $trs 0])"
	    if {[lindex $trs 0] < $lpf} {
		# This is the best match so far!
		set lpf [lindex $trs 0]
		set good_matches {}
	    }
	    set good_matches [concat $good_matches [lindex $trs 1]]
	}
	
	incr cwp
    }

    # puts "---<--- [list $lpf $good_matches]"
    return [list $lpf $good_matches]
}

proc try_special {try pattern cwp mcp lpf} {
    set data {}
    switch [string range [lindex $pattern 0] 1 end] {
	INT
	{
	    if {[regexp {^[0-9]*$} [lindex $try 0]] == 0} {
		return -1
	    }
	    set trs [anchor_left_try \
			 [lrange $try 1 end] \
			 [lrange $pattern 1 end] \
			 $cwp $mcp $lpf]
	    set ncwp [lindex $trs 0]
	    set nmd {}
	    foreach d [lindex $trs 1] {
		lappend nmd [list [lindex $d 0] \
				 [concat [lindex $d 1] \
				      [list [lindex $try 0]]]]
	    }
	    return [list $ncwp $nmd]
	}
	default
	{
	    error "Bad operand type: $p"
	}
    }
    return {-1 {}}
}

# Try-match checks if the user-entered string in $try can possibly
# match the user command template in $entry.  $entry should have the
# form described in the comment above $cmdtab (at the beginning of
# this file).
#
# This function returns a list containing of
#   - A penalty point (0 for a perfect match, higher for each mismatch)
#   - The pretty user name
#   - The TCL command template
#   - A list of arguments to the TCL command template
proc try_match {try entry} {
    set args {}
    set pattern [lindex $entry 0]
    set try [split $try " \t"]
    set trs [anchor_left_try $try $pattern 0 0 1000]
    if {[lindex $trs 0] == 1000} {
	set trs {-1 {}}
    } else {
	set trs [list [lindex $trs 0] \
		     [lindex $entry [expr {[llength $entry] > 2 ? 2 : 0}]] \
		     [lindex $entry 1] \
		     [lindex $trs 1]]
    }
    # puts "try_match returning $trs"
    return $trs
}

# do_command is called whenever the user has entered a command.
# The command is present in $cmd.  If the command is empty, a default
# action is performed.  If the first character is "=", the remainder
# of the line is executed as TCL code.  Otherwise, the command table
# in $cmdtab is used to construct a TCL command to execute.  If no
# match is found, an error message is printed.
proc do_command {cmd} {
    global cmdtab
    global errorCode

    if {[string length $cmd] == 0} {
	perform_command {Foo {default_action}}
	return
    }

    if {[string range $cmd 0 0] == "="} {
	if {[catch {uplevel \#0 [string range $cmd 1 end]} res] == 0} {
	    if {[string length $res] > 0} {
	        print "$res\n"
	    }
	} else {
	    puts "Error: $res"
	    puts "errorCode: $errorCode"
	}
	return
    }

    # FIXME: This is bad for internationalization.
    set cmd [isoify $cmd]

    set best 1000
    set matches {}

    # Collect all matches in $matches.
    foreach entry $cmdtab {
	set score [try_match $cmd $entry]
	set s [lindex $score 0]
	if {$s != -1} {
	    if {$s < $best} {
		set best [lindex $score 0]
		set matches [list [lrange $score 1 end]]
	    } else {
		if {$s == $best} {
		    lappend matches [lrange $score 1 end]
		}
	    }
	}
    }

    # Here, matches looks like this:
    # list of {0: cmd-name 1: cmd-tmpl 2: list of {0: mcp 1: data}}

    # puts "Ok, matches so far: $matches (End matches)"

    set matches [expand_matches $matches]

    # puts "Ok, matches now: $matches (End matches)"

    switch [llength $matches] {
	0
	    {print "Jag frstr inte vad du menar\n"}
	1
	    {perform_command [lindex $matches 0]}
	default
	    {report_ambiguous $matches}
    }
}

proc perform_command {cmd} {
    global senaste
    global errorCode

    # puts "DOING-->[lindex $cmd 1]<--"
    if {[catch {eval [lindex $cmd 1]} res] != 0} {
	puts "Error: $res"
	puts "errorCode: $errorCode"
    }
}

proc report_ambiguous {alts} {
    print "Du kan mena ngot av fljande:\n"
    foreach m $alts {
        print "[lindex $m 0]\n"
    }
}

proc commandline {n} {
    global input_line
    global inhibit_prompt
    global quit_flag

    set inhibit_prompt 0

    switch $n {
	3
	    {return 1}
	10
	    -
	13  
	    {
		print "\n"
		do_command $input_line
	     	set input_line {}
		if {! $inhibit_prompt} {
	     	    print_prompt
		}
	    }
	8
	    -
	127
	    {
		set len [string length $input_line]
		if {$len == 0} {
		    puts -nonewline "\a"
		} else {
		    set input_line \
			[string range $input_line 0 [expr $len - 2]]
		    nilkom_page prompt_print "\008 \008"
		}
	    }
	18
	    {print "\n$input_line"}
	21
	    {
		print "\n"
		set input_line {}
	    }
	default
	    {
		print [format "%c" $n]
		append input_line [format "%c" $n]
	    }
    }
    flush stdout
    nilkom_page sync
    return $quit_flag
}
